@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2023, NVIDIA CORPORATION.
+ * Copyright (c) 2013-2024, NVIDIA CORPORATION.
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation
@@ -42,9 +42,10 @@
 #include "nvidia-modprobe-utils.h"
 #include "pci-enum.h"
 
+
+
 #define NV_DEV_PATH "/dev/"
 #define NV_PROC_MODPROBE_PATH "/proc/sys/kernel/modprobe"
-#define NV_PROC_MODULES_PATH "/proc/modules"
 #define NV_PROC_DEVICES_PATH "/proc/devices"
 
 #define NV_PROC_MODPROBE_PATH_MAX        1024
@@ -53,6 +54,7 @@
 
 #define NV_NVIDIA_MODULE_NAME "nvidia"
 #define NV_PROC_REGISTRY_PATH "/proc/driver/nvidia/params"
+#define NV_PROC_PATH_PREFIX "/proc/driver/nvidia"
 
 #define NV_UVM_MODULE_NAME "nvidia-uvm"
 #define NV_UVM_DEVICE_NAME "/dev/nvidia-uvm"
@@ -89,69 +91,33 @@
 #define NV_MIN(a, b) (((a) < (b)) ? (a) : (b))
 
 /*
- * Just like strcmp(3), except that differences between '-' and '_' are
- * ignored. This is useful for comparing module names, where '-' and '_'
- * are supposed to be treated interchangeably.
- */
-static int modcmp(const char *a, const char *b)
-{
-    int i;
-
-    /* Walk both strings and compare each character */
-    for (i = 0; a[i] && b[i]; i++)
-    {
-        if (a[i] != b[i])
-        {
-            /* ignore differences between '-' and '_' */
-            if (((a[i] == '-') || (a[i] == '_')) &&
-                ((b[i] == '-') || (b[i] == '_')))
-            {
-                continue;
-            }
-
-            break;
-        }
-    }
-
-    /*
-     * If the strings are of unequal length, only one of a[i] or b[i] == '\0'.
-     * If they are the same length, both will be '\0', and the strings match.
-     */
-    return a[i] - b[i];
-}
-
-
-/*
- * Check whether the specified module is loaded by reading
- * NV_PROC_MODULES_PATH; returns 1 if the kernel module is loaded.
+ * Check whether the specified module is loaded by checking if its
+ * initstate file exists; returns 1 if the kernel module is loaded.
  * Otherwise, it returns 0.
  */
 static int is_kernel_module_loaded(const char *nv_module_name)
 {
-    FILE *fp;
-    char module_name[NV_MAX_MODULE_NAME_SIZE];
-    int module_loaded = 0;
+    int i;
+    char init_path[NV_MAX_LINE_LENGTH];
 
-    fp = fopen(NV_PROC_MODULES_PATH, "r");
+    snprintf(init_path, sizeof(init_path),
+             "/sys/module/%s/initstate", nv_module_name);
 
-    if (fp == NULL)
+    /*
+     * modprobe replaces '-' with '_' to generate module name from
+     * file name. Do the same when checking for module presence, by
+     * transforming nv_module_name within init_path[], starting after
+     * "/sys/module/" until "/initstate"
+     */
+    for (i = strlen("/sys/module/");
+         (init_path[i] != '\0') && (init_path[i] != '/');
+         i++)
     {
-        return 0;
+        if (init_path[i] == '-')
+            init_path[i] = '_';
     }
 
-    while (fscanf(fp, "%15s%*[^\n]\n", module_name) == 1)
-    {
-        module_name[15] = '\0';
-        if (modcmp(module_name, nv_module_name) == 0)
-        {
-            module_loaded = 1;
-            break;
-        }
-    }
-
-    fclose(fp);
-
-    return module_loaded;
+    return (access(init_path, R_OK) == 0);
 }
 
 /*
@@ -945,8 +911,16 @@ int nvidia_vgpu_vfio_mknod(int minor_num)
         return 0;
     }
 
-    ret = snprintf(vgpu_dev_name, NV_MAX_CHARACTER_DEVICE_FILE_STRLEN,
-                   NV_VGPU_VFIO_DEVICE_NAME, minor_num);
+    if (minor_num == NV_VGPU_VFIO_CTL_MINOR)
+    {
+        ret = snprintf(vgpu_dev_name, NV_MAX_CHARACTER_DEVICE_FILE_STRLEN, NV_VGPU_VFIO_CTL_NAME);
+    }
+    else
+    {
+        ret = snprintf(vgpu_dev_name, NV_MAX_CHARACTER_DEVICE_FILE_STRLEN,
+                       NV_VGPU_VFIO_DEVICE_NAME, minor_num);
+    }
+
     if (ret <= 0)
     {
         return 0;
@@ -1020,6 +994,19 @@ int nvidia_cap_mknod(const char* cap_file_path, int *minor)
     char name[NV_MAX_CHARACTER_DEVICE_FILE_STRLEN];
     int ret;
 
+    /* Verify the path prefix is an absolute path to the NVIDIA driver /proc */
+    if ((strncmp(cap_file_path, NV_PROC_PATH_PREFIX, strlen(NV_PROC_PATH_PREFIX)) != 0) ||
+        (strstr(cap_file_path, "./") != NULL))
+    {
+        return 0;
+    }
+
+    /* Check if the (real) user has access to the path */
+    if (access(cap_file_path, R_OK) != 0)
+    {
+        return 0;
+    }
+
     ret = nvidia_cap_get_device_file_attrs(cap_file_path, &major, minor, name);
     if (ret == 0)
     {
